"""Utility functions to manage geospatial vector data in geojson format."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.import json
import json

# Given a geojson string, returns a json object after having tested that the input string contains a valid geojson
def geojsonJson(geojson):
    """
    Given a geojson string, returns a json dictionary after having tested that the input string contains a valid geojson
    
    Parameters
    ----------
        geojson : str
            String containing data in geojson format
            
    Returns
    -------
        Json dictionary
        
    Raises
    ------
        Exception if the input string is not in geojson format
    
    """
    j = json.loads(geojson)
    
    if j.get('type', None) != 'FeatureCollection':
        raise Exception('Sorry, input string does not look like GeoJSON')
        
    if type(j.get('features', None)) != list:
        raise Exception('Sorry, input string does not contains GeoJSON features')
        
    return j



# Load a geojson from file, testing that it contains a valid geojson
# Returns a string
def geojsonLoadFile(filepath):
    """
    Load a geojson content from file, testing that it contains valid geojson data
    
    Parameters
    ----------
        filepath : str
            File path of the geojson file to load
            
    Returns
    -------
        File content as a geojson string
        
    Example
    -------
    Load a geojson from file, print the geojson string::
    
        from vois import geojsonUtils
        
        geojson = geojsonUtils.geojsonLoadFile('./data/example.geojson')

        print(geojson)
        
    .. figure:: figures/geojsonLoad.png
       :scale: 100 %
       :alt: Geojson example geojsonLoadFile

       Read a geojson file and print its content

    """

    with open(filepath,"r") as f:
        geojson = f.read()

    j = geojsonJson(geojson)
    return json.dumps(j)



# Given a geojson string, returns the number of features
def geojsonCount(geojson):
    """
    Given a geojson string, returns the number of features
    
    Parameters
    ----------
        geojson : str
            String containing data in geojson format
            
    Returns
    -------
        Integer corresponding to the number of features in the input geojson string
    """
    
    j = geojsonJson(geojson)
    return len(j['features'])


# Given a geojson string, returns the list of the attribute names of the features
def geojsonAttributes(geojson):
    """
    Given a geojson string, returns the list of the attribute names of the features
    
    Parameters
    ----------
        geojson : str
            String containing data in geojson format
            
    Returns
    -------
        List of strings containing the names of the attributes of the features in the input geojson string

    Example
    -------
    Load a geojson from file and print the names of its attributes::
    
        from vois import geojsonUtils
        
        geojson = geojsonUtils.geojsonLoadFile('./data/example.geojson')

        print('Attributes = ', geojsonUtils.geojsonAttributes(geojson))
        
    .. figure:: figures/geojsonAttributes.png
       :scale: 100 %
       :alt: Geojson example geojsonAttributes

       Read a geojson file and print the names of its attributes
       
    """
    
    j = geojsonJson(geojson)
    
    features = j['features']
    attributes = set()
    for f in features:
        attributes.update(f['properties'].keys())
        
    return list(attributes)
    

# Given a geojson string, returns the list of values of the attribute attributeName for all the features
def geojsonAll(geojson, attributeName):
    """
    Given a geojson string, returns the list of values of the attribute attributeName for all the features
    
    Parameters
    ----------
        geojson : str
            String containing data in geojson format
        attributeName : str
            Name of one of the attributes
            
    Returns
    -------
        List containing the values of the attribute for all the features of the input geojson dataset

    Example
    -------
    Load a geojson from file and print all the values of one if its attributes::
    
        from vois import geojsonUtils
        
        geojson = geojsonUtils.geojsonLoadFile('./data/example.geojson')

        print('Values = ', geojsonUtils.geojsonAll(geojson,'ha'))
        
    .. figure:: figures/geojsonAll.png
       :scale: 100 %
       :alt: Geojson example geojsonAll

       Read a geojson file and print the values of one of its attributes for all the features of the dataset
       
    """        
    
    j = geojsonJson(geojson)
    
    features = j['features']
    attributes = set()
    values = [f['properties'][attributeName] if attributeName in f['properties'] else None for f in features]
    return values


# Add a field to a geojson by joining a python dictionary through match with the field named keyname.
# If innerMode is True, the output geojson will only keep the joined features, otherwise all the original features are returned
# Returns the modified geojson
def geojsonJoin(geojson, keyname, addedfieldname, keytovaluedict, innerMode=False):
    """
    Add a field to a geojson by joining a python dictionary through match with the field named keyname.
    If innerMode is True, the output geojson will only keep the joined features, otherwise all the original features are returned
    
    Parameters
    ----------
        geojson : str
            String containing data in geojson format
        keyname : str
            Name of the attribute of the input geojson to use as internal key for the join operation
        addedfieldname : str
            Name of the attribute to add to the input geojson as a result of the join operation
        keytovaluedict : dict
            Dictionary (key-value pairs) to use as joined values. The keys of the <keytovaluedict> are used as foreign keys to match the values of the <keyname> attribute of the input geojson. When a match is found, the attribute <addedfieldname> is added to the corresponding feature having the value read from the <keytovaluedict>
        innerMode : bool, optional
            If innerMode is True, the output geojson will only keep the successfully joined features, otherwise all the original features are returned
            
    Returns
    -------
        a string containing the modified geojson after the join operation
        
    Example
    -------
    Load a geojson from file, print some information on attributes and values of the features, then join the features with a dictionary::
    
        from vois import geojsonUtils
        
        # Load a geojson file
        geojson = geojsonUtils.geojsonLoadFile('./data/example.geojson')

        # Add a new field by joining with a dictionary (with innerMode flag set to True)
        keytovalue = { 37661: 'aaa', 37662: 'bbb'}
        geojsonnew = geojsonUtils.geojsonJoin(geojson,'id', 'value', keytovalue, innerMode=True)
        
        # Print the 'value' attribute values for the joined geojson dataset
        print('Joined values =', geojsonUtils.geojsonAll(geojsonnew,'value'))

    .. figure:: figures/geojsonJoin.png
       :scale: 100 %
       :alt: Geojson example

       Result of the Join operation
       
    """

    j = geojsonJson(geojson)

    # Feature collection to return
    res = {
        "type": "FeatureCollection",
        "features": []
    }

    if 'crs' in j:
        res['crs'] = j['crs']
        
    for f in j['features']:
        if 'properties' in f:
            if keyname in f['properties']:
                key = f['properties'][keyname]
                if key in keytovaluedict:
                    value = keytovaluedict[key]
                    f['properties'][addedfieldname] = value
                    if innerMode:
                        res['features'].append(f)
        if not innerMode:
            res['features'].append(f)
        
    return json.dumps(res)


# Filter a geojson by keeping only the features for which <fieldname> has value <fieldvalue> (fieldvalue can be also a list!)
# Returns the modified geojson
def geojsonFilter(geojson, fieldname, fieldvalue):
    """
    Filter a geojson by keeping only the features for which <fieldname> has value <fieldvalue> (fieldvalue can be also a list)
    
    Parameters
    ----------
        geojson : str
            String containing data in geojson format
        fieldname : str
            Name of one of the attributes of the input geojson
        fieldvalue : single value or list of values
            Comparison value. Only the input features having this value on the <fieldname> attribute are kept in the output geojson returned
            
    Returns
    -------
        a string containing the modified geojson containing only the features that pass the filter operation
    """        

    j = geojsonJson(geojson)

    # Feature collection to return
    res = {
        "type": "FeatureCollection",
        "features": []
    }

    if 'crs' in j:
        res['crs'] = j['crs']
        
    for f in j['features']:
        if 'properties' in f:
            if fieldname in f['properties']:
                fvalue = f['properties'][fieldname]
                if (type(fieldvalue) is list and fvalue in fieldvalue) or (fvalue == fieldvalue):
                    res['features'].append(f)
        
    return json.dumps(res)


"""Interactive page configurator"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Imports
from ipywidgets import widgets
import ipyvuetify as v
import json

# Vois imports
from vois import colors, download
from vois.vuetify import settings, toggle, ColorPicker, sliderFloat, UploadImage, UploadJson, Button, switch, tooltip, iconButton, dialogGeneric, selectSingle
from vois.templates import template1panel, template2panels, template3panels


# Change style of a label
def labelchange(lab, size=14, weight=400, color=settings.color_first, width=None):
    if width is None:
        lab.style_ = 'font-size: %dpx; font-weight: %d; color: %s;'%(size,weight,color)
    else:
        lab.style_ = 'font-size: %dpx; font-weight: %d; color: %s; width: %dpx'%(size,weight,color,int(width))
    
# Creation of a label
def label(text, class_='pa-0 ma-0 mt-1 mr-3', size=14, weight=400, color=settings.color_first, width=None):
    lab = v.Html(tag='div', children=[text], class_=class_)
    labelchange(lab, size, weight, color, width)
    return lab

    
#####################################################################################################################################################
# Interactive page configurator widget
#####################################################################################################################################################
class PageConfigurator(v.Html):

    # Initialization
    def __init__(self, output, **kwargs):

        super().__init__(**kwargs)

        self.debug = widgets.Output()
        
        # Default state
        self.reset_state = {
            'appname':           'My app',
            'title':             'My page',
            'titlecolor':        '#0d856d',
            'titledark':         True,
            'titleheight':       54,
            'titleimageurl':     '',
            'footercolor':       '#a0dcd0',
            'footerdark':        False,
            'footerheight':      30,
            'copyrighttext':     'European Commission - Joint Research Centre',
            'show_back':         True,
            'left_back':         True,
            'show_help':         True,
            'show_credits':      True,
            'logoappurl':        'https://jeodpp.jrc.ec.europa.eu/services/shared/pngs/BDAP_Logo1024transparent.png',
            'logocreditsurl':    'https://jeodpp.jrc.ec.europa.eu/services/shared/pngs/EC-JRC-logo_horizontal_EN_neg_transparent-background.png',
            'logowidth':         40,
            'creditswidth':      120,
            'transition':        'dialog-bottom-transition',
        
            # Custom properties
            'panelsvalue':       0,
            'titledarkvalue':    2,
            'footerlinkedvalue': 0,
            'footerdarkvalue':   2,
            'button_rounded':    False,
            'transitionvalue':   'Bottom'
        }
        
        
        
        self.output = output
        self.spacerX = v.Html(tag='div', style_='width: 10px; height: 0px;')
        self.spacerY = v.Html(tag='div', style_='width:  0px; height: 6px;')
        self.spacer  = v.Html(tag='div', style_='width: 10px; height: 10px;')

        self.card = v.Card(flat=True, width=template1panel.LEFT_WIDTH, min_width=template1panel.LEFT_WIDTH, max_width=template1panel.LEFT_WIDTH,
                           height='400px', class_='pa-2 pt-4 ma-0', style_='overflow: auto;')

        # Widgets
        self.labelwidth   = 96
        self.togglewidth  = 46
        self.paddingrow   = 1
        self.biglabelsize = 15
        
        self.appname   = v.TextField(label='Application name:', autofocus=False, v_model=None, dense=False, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3 mr-3")
        self.buttOpen  = iconButton.iconButton(icon='mdi-folder-open',  onclick=self.onOpen,  tooltip='Load state from file',                      margins='pa-0 ma-0 mt-3 mr-2')
        self.buttSave  = iconButton.iconButton(icon='mdi-content-save', onclick=self.onSave,  tooltip='Save current state to file',                margins='pa-0 ma-0 mt-3 mr-2')
        self.buttCode  = iconButton.iconButton(icon='mdi-file-code',    onclick=self.onCode,  tooltip='Generate and download page code in Python', margins='pa-0 ma-0 mt-3 mr-2')
        self.buttReset = iconButton.iconButton(icon='mdi-backspace',    onclick=self.onReset, tooltip='Reset page to default state',               margins='pa-0 ma-0 mt-3 mr-2')
        
        self.cardappname = v.Card(flat=True, width=template1panel.LEFT_WIDTH-46, max_width=template1panel.LEFT_WIDTH-46,
                                  children=[widgets.HBox([self.appname, self.buttOpen.draw(), self.buttSave.draw(), self.buttCode.draw(), self.buttReset.draw()])])
        
        self.pagetitle  = v.TextField(label='Page title:', autofocus=False, v_model=None, dense=False, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3 mr-3", style_='width: 246px; min_width: 246px;')
        self.transition = selectSingle.selectSingle('Open animation:', ['None', 'Dialog', 'Bottom'], selection='Bottom', clearable =False, width=120, onchange=self.transitionChange)
        
        self.cardpagetitle = v.Card(flat=True, width=template1panel.LEFT_WIDTH, max_width=template1panel.LEFT_WIDTH,
                                    children=[widgets.HBox([self.pagetitle, self.transition.draw()])])
        
        self.appname.on_event(  'change',      self.appnameChange)
        self.appname.on_event(  'click:clear', self.appnameClear)
        self.pagetitle.on_event('change',      self.pagetitleChange)
        self.pagetitle.on_event('click:clear', self.pagetitleClear)
        
        
        self.panelsLabel  = label('Page format: ', size=self.biglabelsize, weight=500, width=self.labelwidth)
        self.togglePanels = toggle.toggle(self.reset_state['panelsvalue'],
                                          ['1', '2', '3'],
                                          tooltips=['Page with 1 left panel', 'Page with 2 panels: left and bottom', 'Page with 3 panels: left, bottom and right'],
                                          dark=True, onchange=self.onSelectedTemplate, row=True, width=self.togglewidth, justify='start', paddingrow=self.paddingrow, tile=True)

        
        self.titlecolor  = ColorPicker(color=settings.color_first,  width=50, height=30, rounded=False, on_change=self.titlecolorChange,  offset_x=True, offset_y=False, color_theory_popup=True)        
        self.footercolor = ColorPicker(color=settings.color_second, width=50, height=30, rounded=False, on_change=self.footercolorChange, offset_x=True, offset_y=False, color_theory_popup=True)
        
        self.rounded = switch.switch(self.reset_state['button_rounded'], 'Rounded buttons', inset=True, dense=True, onchange=self.onroundedChange)
        
        self.footerlinked = toggle.toggle(self.reset_state['footerlinkedvalue'], ['', '', '', ''], dark=True, icons=['mdi-link-off', 'mdi-link-variant', 'mdi-link-variant-minus', 'mdi-link-variant-plus'], outlined=False,
                                          tooltips=['Free selection of footer color', 'Footer color is the complementary of the title color',
                                                    'Footer color is a darker version of the title color', 'Footer color is a lighter version of the title color'],
                                          onchange=self.footerlinkedChange, row=True, width=self.togglewidth-20, height=30, justify='start', paddingrow=self.paddingrow, tile=True)
        
        self.titledark = toggle.toggle(self.reset_state['titledarkvalue'], ['', '', ''], dark=True, icons=['mdi-alpha-w-box-outline', 'mdi-alpha-b-box-outline', 'mdi-auto-fix'],
                                       tooltips=['Display text in white color on the title bar', 'Display text in black color on the title bar', 'Automatically select text color for the title bar'],
                                       onchange=self.titledarkChange, row=True, width=self.togglewidth, justify='start', paddingrow=self.paddingrow, tile=True)

        self.footerdark = toggle.toggle(self.reset_state['footerdarkvalue'], ['', '', ''], dark=True, icons=['mdi-alpha-w-box-outline', 'mdi-alpha-b-box-outline', 'mdi-auto-fix'],
                                        tooltips=['Display text in white color on the footer bar', 'Display text in black color on the footer bar', 'Automatically select text color for the footer bar'],
                                        onchange=self.footerdarkChange, row=True, width=self.togglewidth, justify='start', paddingrow=self.paddingrow, tile=True)
        
        self.titleheight = sliderFloat.sliderFloat(self.reset_state['titleheight'], text='Title bar height:', minvalue=20.0, maxvalue=180.0, maxint=160, showpercentage=False, decimals=0,
                                                   labelwidth=self.labelwidth-10, sliderwidth=150, resetbutton=True, showtooltip=True, onchange=self.titleheightChange)
        
        self.footerheight = sliderFloat.sliderFloat(self.reset_state['footerheight'], text='Footer bar height:', minvalue=16.0, maxvalue=80.0, maxint=64, showpercentage=False, decimals=0,
                                                    labelwidth=self.labelwidth-10, sliderwidth=150, resetbutton=True, showtooltip=True, onchange=self.footerheightChange)
        
        self.upload = UploadImage.UploadImage(self.output, width=620)
        
        self.titleimageurl  = Button('Select background image for the title bar', color_selected=settings.color_first, dark=settings.dark_mode, 
                                     text_weight=450, on_click=self.titleimageurlClick, width=template1panel.LEFT_WIDTH-16, height=40,
                                     tooltip='Click to select an image to use as background on the title bar', selected=True, rounded=self.reset_state['button_rounded'])

        self.logoappurl     = Button('Select image for the application logo', color_selected=settings.color_first, dark=settings.dark_mode, 
                                     text_weight=450, on_click=self.logoappurlClick, width=template1panel.LEFT_WIDTH-16, height=40,
                                     tooltip='Click to select an image to use as application logo on the left side of the title bar', selected=True, rounded=self.reset_state['button_rounded'])
        
        self.logowidth = sliderFloat.sliderFloat(self.reset_state['logowidth'], text='Application logo width:', minvalue=20.0, maxvalue=200.0, maxint=180, showpercentage=False, decimals=0,
                                                 labelwidth=self.labelwidth-10, sliderwidth=150, resetbutton=True, showtooltip=True, onchange=self.logowidthChange)
        
        self.logocreditsurl = Button('Select image for the credits logo', color_selected=settings.color_first, dark=settings.dark_mode, 
                                     text_weight=450, on_click=self.logocreditsurlClick, width=template1panel.LEFT_WIDTH-16, height=40,
                                     tooltip='Click to select an image to use as credits logo on the right side of the title bar', selected=True, rounded=self.reset_state['button_rounded'])
        
        self.creditswidth = sliderFloat.sliderFloat(self.reset_state['creditswidth'], text='Credits logo width:', minvalue=20.0, maxvalue=300.0, maxint=280, showpercentage=False, decimals=0,
                                                 labelwidth=self.labelwidth-10, sliderwidth=150, resetbutton=True, showtooltip=True, onchange=self.creditswidthChange)

        self.show_back    = switch.switch(self.reset_state['show_back'],    'Show Back button',        inset=True, dense=True, onchange=self.onshow_backChange)
        self.left_back    = switch.switch(self.reset_state['left_back'],    'Back button on the Left', inset=True, dense=True, onchange=self.onleft_backChange)
        self.show_help    = switch.switch(self.reset_state['show_help'],    'Show Help button',        inset=True, dense=True, onchange=self.onshow_helpChange)
        self.show_credits = switch.switch(self.reset_state['show_credits'], 'Show Credits logo',       inset=True, dense=True, onchange=self.onshow_creditsChange)
        
        self.copyrighttext = v.TextField(label='Copyright text:', autofocus=False, v_model=None, dense=False, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3")
        self.copyrighttext.on_event('change',      self.copyrighttextChange)
        self.copyrighttext.on_event('click:clear', self.copyrighttextClear)
        
        self.card.children = [widgets.VBox([
                                 self.cardappname,
                                 self.cardpagetitle,
                                 self.spacerY,
                                 widgets.HBox([self.panelsLabel, self.togglePanels.draw()]),
                                 self.spacerY,
                                 self.spacerY,
                                 widgets.HBox([label('Title bar color:',  color='black', width=self.labelwidth), self.titlecolor, self.titlecolor.ctpopup.draw(), self.spacerX, label('Link:', color='black', width=30), self.footerlinked.draw()]),
                                 self.spacerY,
                                 widgets.HBox([label('Footer color:',     color='black', width=self.labelwidth), self.footercolor, self.footercolor.ctpopup.draw(), self.spacerX, self.rounded.draw()]),
                                 self.spacerY,
                                 widgets.HBox([label('Title bar text:',   color='black', width=self.labelwidth), self.titledark.draw()]),
                                 self.spacerY,
                                 widgets.HBox([label('Footer bar text:',  color='black', width=self.labelwidth), self.footerdark.draw()]),
                                 self.spacerY,
                                 self.titleheight.draw(),
                                 self.spacerY,
                                 self.footerheight.draw(),
                                 self.spacerY,
                                 self.titleimageurl,
                                 self.spacerY,
                                 self.logoappurl,
                                 self.logowidth.draw(),
                                 self.spacerY,
                                 self.logocreditsurl,
                                 self.creditswidth.draw(),
                                 self.spacerY,
                                 widgets.HBox([tooltip.tooltip('Add a back button in the title bar to close the page',self.show_back.draw()),
                                               tooltip.tooltip('Position the back button on the left side of the title bar',self.left_back.draw())]),
                                 widgets.HBox([tooltip.tooltip('Add a help button in the title bar',self.show_help.draw()),
                                               tooltip.tooltip('Display credits logo on the right side of the title bar',self.show_credits.draw())]),
                                 self.spacerY,
                                 self.copyrighttext,
                                ])
                             ]
        
        # Set v.Html members
        self.tag = 'div'
        self.children = [self.card]
        
        # Initial page
        self.saveFilename = None
        self.page = None
        self.onSelectedTemplate(self.reset_state['panelsvalue'])
        
        # Set the initial state
        self.page.state = self.reset_state
        
        # Initialise widgets
        self.appname.v_model       = self.page.appname
        self.pagetitle.v_model     = self.page.title
        self.copyrighttext.v_model = self.page.copyrighttext
        
        
    # Load state from file
    def onOpen(self):
        uj = UploadJson.UploadJson(self.output, onOK=self.onSelectedState, required_attributes=['appname', 'title'], attributes_width=80,
                                   color=self.page.titlecolor, dark=self.page.titledark)
        uj.show()
        
        
    # Called when the UploadJson dialog-box is closed with the OK button
    def onSelectedState(self, state):

        with self.debug:
            
            # Set the page state (does a refresh if the page is open)
            self.page.state = state
        
            # Set the state of the confuguration widgets
            self.togglePanels.value    = state['panelsvalue']
            self.appname.v_model       = self.page.appname
            self.pagetitle.v_model     = self.page.title
            self.titlecolor.color      = self.page.titlecolor
            self.titledark.value       = state['titledarkvalue']
            self.titleheight.value     = self.page.titleheight
            self.footercolor.color     = self.page.footercolor
            self.footerlinked.value    = state['footerlinkedvalue']
            self.footerdark.value      = state['footerdarkvalue']
            self.footerheight.value    = self.page.footerheight
            self.show_back.value       = self.page.show_back
            self.left_back.value       = self.page.left_back
            self.show_help.value       = self.page.show_help
            self.show_credits.value    = self.page.show_credits
            self.logowidth.value       = self.page.logowidth
            self.creditswidth.value    = self.page.creditswidth
            self.copyrighttext.v_model = self.page.copyrighttext
            self.rounded.value         = state['button_rounded']
            if 'transitionvalue' in state:
                self.transition.value = state['transitionvalue']
            else:
                self.transition.value = 'Bottom'
        
        
    # Save current state to file
    def onSave(self):
        self.saveFilename = v.TextField(label='File name:', autofocus=True, v_model=self.page.appname + '_' + self.page.title, dense=False, color=self.page.titlecolor, clearable=False, class_="pa-0 ma-0 ml-3 mt-3 mr-3")
        dialogGeneric.dialogGeneric(title='Save and download current state...' , on_ok=self.onDoSaveState, 
                                    text='   ', color=self.page.titlecolor, dark=self.page.titledark,
                                    titleheight=40, width=600,
                                    show=True, addclosebuttons=True, addokcancelbuttons=True,
                                    fullscreen=False, content=[self.saveFilename], output=self.output)
        
    
    # Effective save and download of the current state
    def onDoSaveState(self):
        
        filename = self.saveFilename.v_model
        if filename[-5:] != '.json':
            filename += '.json'
        
        # Read the state from the page
        state = self.page.state
    
        # Add additional states from the PageConfigurator
        state['panelsvalue']       = self.togglePanels.value
        state['titledarkvalue']    = self.titledark.value
        state['footerlinkedvalue'] = self.footerlinked.value
        state['footerdarkvalue']   = self.footerdark.value
        state['button_rounded']    = self.rounded.value
        state['trainsitionvalue']  = self.transition.value
    
        # Convert to string and download
        txt = json.dumps(state)
        download.downloadText(txt, fileName=filename)
            
    
    # Save Python page code
    def onCode(self):
        self.saveFilename = v.TextField(label='Class name:', autofocus=True, v_model=self.page.appname + '_' + self.page.title, dense=False, color=self.page.titlecolor, clearable=False, class_="pa-0 ma-0 ml-3 mt-3 mr-3")
        dialogGeneric.dialogGeneric(title='Save and download Python code...' , on_ok=self.onDoSaveCode, 
                                    text='   ', color=self.page.titlecolor, dark=self.page.titledark,
                                    titleheight=40, width=600,
                                    show=True, addclosebuttons=True, addokcancelbuttons=True,
                                    fullscreen=False, content=[self.saveFilename], output=self.output)
    
        
    # Effective save and download of the Python code
    def onDoSaveCode(self):
        
        self.onDoSaveState()   # Saves <filename>.json

        with self.debug:
            jsonfile = self.saveFilename.v_model
            if jsonfile[-5:] != '.json':
                jsonfile += '.json'
                
            classname = self.saveFilename.v_model
            pythonfile = classname
            pythonfile += '.py'

            notebookfile = classname
            notebookfile += '.ipynb'

            # Selection of the base class
            if self.togglePanels.value == 0:
                baseclass = 'template1panel'
            elif self.togglePanels.value == 1:
                baseclass = 'template2panels'
            else:
                baseclass = 'template3panels'

            # Code generation
            txt = '''from vois.vuetify import settings
settings.color_first    = '%s'
settings.color_second   = '%s'
settings.button_rounded = %s
settings.dark_mode      = %s

from ipywidgets import widgets, HTML, Layout
from IPython.display import display
import ipyvuetify as v
import json

from vois import cssUtils
from vois.vuetify import switch, Button, dialogMessage
from vois.templates import %s


output = widgets.Output(layout=Layout(width='0px', height='0px'))
display(output)

cssUtils.allSettings(output)
cssUtils.switchFontSize(output,14)

# Derived page class
class %s(%s.%s):
    
    # Initialisation
    def __init__(self, output, **kwargs):
        super().__init__(output=output, on_logoapp=self.on_logoapp, on_help=self.on_help, on_credits=self.on_credits, **kwargs)

    # Clicked the application logo
    def on_logoapp(self):
        dialogMessage.dialogMessage(title='Info on ' + self.appname, titleheight=36,
                                    text='Text to customise for info on the application<br>Add text here or open a PDF file',
                                    addclosebuttons=True, show=True, width=400, output=self.output)
        
    # Clicked the 'Help' button
    def on_help(self):
        dialogMessage.dialogMessage(title='Help', titleheight=36,
                                    text='Text to customise for the application help<br>Add text here or open a PDF file',
                                    addclosebuttons=True, show=True, width=400, output=self.output)
    
    # Clicked the credits logo
    def on_credits(self):
        dialogMessage.dialogMessage(title='Credits for ' + self.appname, titleheight=36,
                                    text='Text to customise for the credits info<br>Add text here or open a PDF file',
                                    addclosebuttons=True, show=True, width=400, output=self.output)
                                    
    # Create the content of the Main panel
    def createMain(self):
        super().createMain()

    # Create sample widgets on the Left panel
    def createLeft(self):
        super().createLeft()

        # Create sample widgets
        spacer  = v.Html(tag='div', style_='width: 10px; height: 10px;')
        lab = v.Html(tag='div', children=['Place here your widgets:'], class_='pa-0 ma-0 ml-4 mt-4 mb-3')
        sw  = switch.switch(True, 'Sample switch', inset=True, dense=True, onchange=None)
        b   = Button('Sample button with icon and tooltip', on_click=None, width=360, height=42,
                     tooltip='Tooltip for button', selected=True, icon='mdi-cogs')
        
        # Add widgets to Left card
        self.cardLeft.children = [widgets.VBox([lab, 
                                                sw.draw(), 
                                                spacer,
                                                widgets.HBox([spacer, b])
                                               ])]

# Create an instance of mypage
p = %s(output)
p.create()

# Read the state of the page from the .json file
with open('%s') as f:
    p.state = json.load(f)

# Open the page
p.open()'''%(self.titlecolor.color, self.footercolor.color, str(self.rounded.value), str(self.page.titledark),
             baseclass, classname, baseclass, baseclass, classname, jsonfile)
        
            # Convert to string and download .py file
            download.downloadText(txt, fileName=pythonfile)

            lines = txt.split('\n')
            lines = ",\n".join(["\"%s\\n\""%x for x in lines])

            txt = '''{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "65d32196-829e-41b9-a4e4-bdb6381998e9",
   "metadata": {},
   "outputs": [],
   "source": [
    %s
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python [interapro_env]",
   "language": "python",
   "name": "conda-env-interapro_env-py"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.6"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
'''%lines
        
            # Download .ipynb file
            download.downloadText(txt, fileName=notebookfile)

        
    # Reset page to initial state
    def onReset(self):
        self.onSelectedState(self.reset_state)
            

    # Forced close
    def on_force_close(self, *args):
        self.output.clear_output()    # V.I.!!! removes double draw of popups!!!

        
    # Selection of 1, 2 or 3 panels template
    def onSelectedTemplate(self, index):

        # Default state (back button on the left and JEODPP logo as application logo
        statusdict = {
            'left_back'  : True,
            'logoappurl' : 'https://jeodpp.jrc.ec.europa.eu/services/shared/pngs/BDAP_Logo1024transparent.png'
        }
        
        # Read the state and close the current page
        if self.page is not None:
            statusdict = self.page.state
            self.page.close()

        # Create the instance of the page with the requested number of panels
        if index == 0:
            self.page = template1panel.template1panel(self.output, onclose=self.on_force_close)
        elif index == 1:
            self.page = template2panels.template2panels(self.output, onclose=self.on_force_close)
        else:
            self.page = template3panels.template3panels(self.output, onclose=self.on_force_close)
            
        self.page.customButtonAdd('mdi-delete', 'Click here to force page closing', self.on_force_close)
            
        # Create the page widgets and display the PageConfigurator in the left panel
        self.page.create()
        self.page.cardLeft.children = [self]

        # Set the state
        self.page.state = statusdict

        self.page.toggleBasemap.dark            = self.page.titledark
        self.page.toggleBasemap.colorselected   = self.page.titlecolor
        self.page.toggleBasemap.colorunselected = self.page.footercolor
        
        self.titleimageurl.color_selected  = self.page.titlecolor
        self.titleimageurl.dark            = self.page.titledark
        
        self.logoappurl.color_selected     = self.page.titlecolor
        self.logoappurl.dark               = self.page.titledark

        self.logocreditsurl.color_selected = self.page.titlecolor
        self.logocreditsurl.dark           = self.page.titledark
        
        # Open the page with minimal transition
        self.page.transition = 'dialog-top-transition'
        self.page.open()
        if 'transition' in statusdict:
            self.page.transition = statusdict['transition']
        else:
            self.page.transition = 'dialog-bottom-transition'

        # Dimension the left card
        self.card.height = self.card.min_height = self.card.max_height = self.page.height

        
    # Change the transition mode
    def transitionChange(self):
        if self.transition.value == 'None':
            self.page.transition = 'dialog-top-transition'
        elif self.transition.value == 'Dialog':
            self.page.transition = 'dialog-transition'
        else:
            self.page.transition = 'dialog-bottom-transition'
        

    # Change of the titlecolor property
    def titlecolorChange(self):
        color = self.titlecolor.color

        # page color
        self.page.titlecolor = color
        
        # widgets color
        self.appname.color                    = color
        self.pagetitle.color                  = color
        self.transition.color                 = color
        self.buttOpen.color                   = color
        self.buttSave.color                   = color
        self.buttCode.color                   = color
        self.buttReset.color                  = color
        self.page.toggleBasemap.colorselected = color
        self.togglePanels.colorselected       = color
        self.titledark.colorselected          = color
        self.footerdark.colorselected         = color
        self.footerlinked.colorselected       = color
        self.titleheight.slider.color         = color
        self.footerheight.slider.color        = color
        self.titleimageurl.b.color            = color
        self.logoappurl.b.color               = color
        self.logocreditsurl.b.color           = color
        self.logowidth.slider.color           = color
        self.creditswidth.slider.color        = color
        self.copyrighttext.color              = color
        self.show_back.switch.color           = color
        self.left_back.switch.color           = color
        self.show_help.switch.color           = color
        self.show_credits.switch.color        = color
        self.rounded.switch.color             = color
        self.titledarkChange(self.titledark.value)
                
        # labels color
        labelchange(self.panelsLabel, size=self.biglabelsize, weight=500, color=color, width=self.labelwidth)
        
        # Auto calculate footer color
        if self.footerlinked.index == 1:     # complementary
            color = self.titlecolor.color
            colortuple = colors.string2rgb(color)
            complementary = colors.complementaryColor(colortuple)
            self.footercolor.color = colors.rgb2hex(complementary)

        elif self.footerlinked.index == 2:   # monochrome darker
            color = self.titlecolor.color
            colortuple = colors.string2rgb(color)
            monochrome = colors.monochromaticColor(colortuple, -0.3)
            self.footercolor.color = colors.rgb2hex(monochrome)

        elif self.footerlinked.index == 3:   # monochrome lighter
            color = self.titlecolor.color
            colortuple = colors.string2rgb(color)
            monochrome = colors.monochromaticColor(colortuple, 0.3)
            self.footercolor.color = colors.rgb2hex(monochrome)
            
        
    # Change of the footercolor property
    def footercolorChange(self):
        color = self.footercolor.color
        
        # page color
        self.page.footercolor = color

        # widgets color
        self.page.toggleBasemap.colorunselected = color
        self.togglePanels.colorunselected       = color
        self.titledark.colorunselected          = color
        self.footerdark.colorunselected         = color
        self.footerlinked.colorunselected       = color
        self.footerdarkChange(self.footerdark.value)
                
            
    # Button rounded flag
    def onroundedChange(self, flag):
        self.titleimageurl.rounded  = flag
        self.logoappurl.rounded     = flag
        self.logocreditsurl.rounded = flag
    
        
    # Change of the titledark property
    def titledarkChange(self, index):

        # White text color
        if index == 0:
            flag = True

        # Black text color
        elif index == 1:
            flag = False
        
        # Auto color depending on title bar color
        else:
            color = self.titlecolor.color
            colortuple = colors.string2rgb(color)
            if colors.isColorDark(colortuple):
                flag = True
            else:
                flag = False

        self.page.titledark          = flag
        self.togglePanels.dark       = flag
        self.titledark.dark          = flag
        self.footerdark.dark         = flag
        self.footerlinked.dark       = flag
        self.titleimageurl.dark      = flag
        self.logoappurl.dark         = flag
        self.logocreditsurl.dark     = flag
        self.page.toggleBasemap.dark = flag
        self.footercolor.dark_text   = flag
                
            
    # Change of the footerdark property
    def footerdarkChange(self, index):

        # White text color
        if index == 0:
            self.page.footerdark = True
        
        # Black text color
        elif index == 1:
            self.page.footerdark = False
        
        # Auto color depending on title bar color
        else:
            color = self.footercolor.color
            colortuple = colors.string2rgb(color)
            if colors.isColorDark(colortuple):
                self.page.footerdark = True
            else:
                self.page.footerdark = False
                
                
    # Change of the footerlinked toggle
    def footerlinkedChange(self, index):
        self.footercolor.disabled = index > 0
        self.titlecolorChange()
        
        
    # Change of the title bar height
    def titleheightChange(self, height):
        self.page.titleheight = int(height)
    
    
    # Change of the footer bar height
    def footerheightChange(self, height):
        self.page.footerheight = int(height)
        
        
    # Change application name
    def appnameChange(self, *args):
        if self.appname.v_model is None: name = ''
        else:                            name = self.appname.v_model
        self.page.appname = name

    def appnameClear(self, *args):
        self.page.appname = ''
            
    # Change page title
    def pagetitleChange(self, *args):
        if self.pagetitle.v_model is None: name = ''
        else:                              name = self.pagetitle.v_model
        self.page.title = name

    def pagetitleClear(self, *args):
        self.page.title = ''
            
            
    # Click on the button to open the dialog to select the title bar image
    def titleimageurlClick(self, *args):
        self.upload.title   = 'Select background image for the title bar'
        self.upload.onOK    = self.titleimageurlSelected
        self.upload.color   = self.titlecolor.color
        self.upload.u.color = self.titlecolor.color
        self.upload.dark    = self.page.titledark
        self.upload.show()
        
    # Selection of the title bar image
    def titleimageurlSelected(self, imageurl):
        self.page.titleimageurl = imageurl
        
        
    # Click on the button to open the dialog to select the logoappurl
    def logoappurlClick(self, *args):
        self.upload.title   = 'Select image for the application logo'
        self.upload.onOK    = self.logoappurlSelected
        self.upload.color   = self.titlecolor.color
        self.upload.u.color = self.titlecolor.color
        self.upload.dark    = self.page.titledark
        self.upload.show()
        
    # Selection of the logoappurl
    def logoappurlSelected(self, imageurl):
        self.page.logoappurl = imageurl
        
        
    # Click on the button to open the dialog to select the logocreditsurl
    def logocreditsurlClick(self, *args):
        self.upload.title   = 'Select image for the credits logo'
        self.upload.onOK    = self.logocreditsurlSelected
        self.upload.color   = self.titlecolor.color
        self.upload.u.color = self.titlecolor.color
        self.upload.dark    = self.page.titledark
        self.upload.show()
        
    # Selection of the logocreditsurl
    def logocreditsurlSelected(self, imageurl):
        self.page.logocreditsurl = imageurl
        
    # Change logo width
    def logowidthChange(self, width):
        self.page.logowidth = int(width)
    
    # Change credits width
    def creditswidthChange(self, width):
        self.page.creditswidth = int(width)
        

    # Change copyright text
    def copyrighttextChange(self, *args):
        if self.copyrighttext.v_model is None: name = ''
        else:                                  name = self.copyrighttext.v_model
        self.page.copyrighttext = name

    def copyrighttextClear(self, *args):
        self.page.copyrighttext = ''
        
    
    # show_back change
    def onshow_backChange(self, flag):
        self.page.show_back = flag

    # left_back change
    def onleft_backChange(self, flag):
        self.page.left_back = flag

    # show_help change
    def onshow_helpChange(self, flag):
        self.page.show_help = flag

    # show_credits change
    def onshow_creditsChange(self, flag):
        self.page.show_credits = flag
        


"""Widget to select among alternative options using a list of buttons displayed horizontally or vertically."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v

try:
    from . import settings
    # from . import Button
except:
    import settings
    # import Button

from vois.vuetify import Button

#####################################################################################################################################################
# Toggle buttons control
#####################################################################################################################################################
class toggle():
    """
    Widget to select among alternative options using a list of buttons displayed horizontally or vertically.
        
    Parameters
    ----------
    index : int
        Index of the selected option at start (from 0 to len(labels)-1)
    labels : list of strings
        Strings to be displayed as text of the options
    tooltips : list of strings, optional
        Tooltip text for the options (default is the empty list)
    icons : list of strings, optional
        Icons for the options (default is the empty list)
    onchange : function, optional
        Python function to call when the user clicks on one of the buttons. The function will receive a parameter of type int containing the index of the clicked button, from 0 to len(labels)-1
    row : bool, optional
        Flag to display the buttons horizontally or vertically (default is True)
    width : int, optional
        Width in pixels of the buttons (default is 150)
    height : int, optional
        Height in pixels of the buttons (default is 36)
    justify : str, optional
        In case of horizontal placement, applies the justify-content css property. Available options are: start, center, end, space-between and space-around.
    rounded : bool, optional
        Flag to display the buttons with a rounded shape (default is the button_rounded flag defined in the settings.py module)
    outlined : bool, optional
        Flag to display the buttons as outlined (default is False)
    colorselected : str, optional
        Color used for the buttons when they are selected (default is settings.color_first)
    colorunselected : str, optional
        Color used for the buttons when they are not selected (default is settings.color_second)
    dark : bool, optional
        Flag that controls the color of the text and icons in foreground (if True, the text and icons will be displayed in white, otherwise in black)
    paddingrow : int, optional
        Horizontal padding among toggle buttons (1 unit means 4 pixels). Default is 1
    paddingcol : int, optional
        Vertical padding among toggle buttons (1 unit means 4 pixels). Default is 2
    tile : bool, optional
        Flag to remove the buttons small border (default is False)
    large : bool, optional
        Flag that sets the large version of the button (default is False)
    xlarge : bool, optional
        Flag that sets the xlarge version of the button (default is False)
    small : bool, optional
        Flag that sets the small version of the button (default is False)
    xsmall : bool, optional
        Flag that sets the xsmall version of the button (default is False)

    Example
    -------
    Creation and display of a widget for the selection among 3 options::
        
        from vois.vuetify import toggle
        from ipywidgets import widgets
        from IPython.display import display
        
        output = widgets.Output()

        def onchange(index):
            with output:
                print(index)

        t = toggle.toggle(0, ['Option 1', 'Option 2', 'Option 3'], tooltips=['Tooltip for option 1'], 
                          onchange=onchange, row=False, width=150, rounded=False)

        display(t.draw())
        display(output)


    .. figure:: figures/toggle.png
       :scale: 100 %
       :alt: toggle widget

       Toogle widget for selecting alternative options using buttons.
   """

    # Initialization
    def __init__(self, index, labels, tooltips=[], icons=[], onchange=None,
                 row=True, width=150, height=36, justify='space-between', rounded=settings.button_rounded, outlined=False,
                 colorselected=settings.color_first, colorunselected=settings.color_second, dark=settings.dark_mode,
                 paddingrow=1, paddingcol=2, tile=False, small=False, xsmall=False, large=False, xlarge=False):
        
        self.index      = index    # Index of the selected button
        self.labels     = labels
        self.tooltips   = tooltips
        self.icons      = icons
        self.onchange   = onchange
        self.row        = row
        self.width      = width
        self.height     = height
        self.justify    = justify
        self.rounded    = rounded
        self.outlined   = outlined
        self._colorselected   = colorselected
        self._colorunselected = colorunselected
        self._dark            = dark
        self.paddingrow       = paddingrow
        self.paddingcol       = paddingcol
        self.tile   = tile
        self.small  = small
        self.xsmall = xsmall
        self.large  = large
        self.xlarge = xlarge
        
        self.__createButtons()
        
        
    # Create the toggle buttons
    def __createButtons(self):

        iconcolor = 'black'
        if self._dark:
            iconcolor = 'white'
                
        self.buttons = []
        i = 0
        for label in self.labels:
            tooltip = ''
            if i < len(self.tooltips):
                tooltip = self.tooltips[i]

            icon = None
            if i < len(self.icons):
                icon = self.icons[i]
                
                
            if self.row: c = "pa-0 ma-0 mr-%d"%self.paddingrow
            else:        c = "pa-0 ma-0 mb-%d"%self.paddingcol
                    
            b = Button(label, dark=self._dark, class_=c, small=self.small, xsmall=self.xsmall, large=self.large, xlarge=self.xlarge,
                       icon=icon, icon_color=iconcolor,
                       onclick=self.__internal_onchange, argument=i, width=self.width, height=self.height, tooltip=tooltip, selected=(i==self.index),
                       rounded=self.rounded, tile=self.tile, outlined=self.outlined, colorselected=self._colorselected, colorunselected=self._colorunselected)
            self.buttons.append(b)
            i += 1

        if self.row: self.group = v.Row(class_="pa-0 ma-0", justify=self.justify, children=[x.draw() for x in self.buttons], style_="overflow: hidden;")
        else:        self.group = v.Col(cols=12, class_="pa-0 ma-0", children=[x.draw() for x in self.buttons], style_="overflow: hidden;")


    # Get the active button
    @property
    def value(self):
        """
        Get/Set the active option index.
        
        Returns
        --------
        index : int
            Index of the selected option (from 0 to len(labels)-1)

        Example
        -------
        Programmatically select one of the options and print the index of the selected option::
            
            t.value = 2
            print(t.value)
        
        """
        return self.index
   
    
    # Set the active button
    @value.setter
    def value(self, index):
        if index >= 0 and index < len(self.buttons):
            self.buttons[self.index].selected = False
            self.index = index
            self.buttons[self.index].selected = True
            if self.onchange:
                self.onchange(self.index)
            

    # Get/Set the color of the selected button
    @property
    def colorselected(self):
        """
        Get/Set the color of the selected button.
        
        Returns
        --------
        color : str
            Color of the selected button

        Example
        -------
        Programmatically change the color of the selected button::
            
            t.colorselected = '#ff0000'
            print(t.colorselected)
        
        """
        return self._colorselected
   
    
    @colorselected.setter
    def colorselected(self, color):
        for b in self.buttons:
            b.color_selected = color
            if b._selected:
                b.b.color = color
                
                
    # Get/Set the color of the unselected button
    @property
    def colorunselected(self):
        """
        Get/Set the color of the unselected buttons.
        
        Returns
        --------
        color : str
            Color of the unselected buttons

        Example
        -------
        Programmatically change the color of the unselected buttons::
            
            t.colorunselected = '#ff0000'
            print(t.colorunselected)
        
        """
        return self._colorunselected
   
    
    @colorunselected.setter
    def colorunselected(self, color):
        for b in self.buttons:
            b.color_unselected = color
            if not b._selected:
                b.b.color = color
                
                
    # Get/Set the dark mode
    @property
    def dark(self):
        """
        Get/Set the dark mode of the buttons.
        
        Returns
        --------
        flag : bool
            If True the text and the icons are displayed in white color, otherwise in black color

        Example
        -------
        Programmatically change the dark mode::
            
            t.dark = Fals
            print(t.dark)
        
        """
        return self._dark
   
    
    @dark.setter
    def dark(self, flag):
        self._dark = flag
        
        for b in self.buttons:
            b.dark = self._dark
                
                
    # Manage onchange event
    def __internal_onchange(self, index):
        self.buttons[self.index].selected = False
        self.index = index
        self.buttons[self.index].selected = True
        if self.onchange:
            self.onchange(self.index)
    
    # Returns the vuetify object to display
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Row or v.Col widget)"""
        return self.group




"""Utility functions for the creation of interactive maps using ipyleaflet Map."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import pandas as pd
import numpy as np
import json
import random
from pathlib import Path

from ipywidgets import widgets, Layout, HTML, Label
from ipyleaflet import Map, basemaps, GeoJSON, Popup, SearchControl, WidgetControl, LegendControl, ScaleControl, FullScreenControl

try:
    from . import colors
    from . import geojsonUtils
except:
    import colors
    import geojsonUtils


###########################################################################################################################################################################
# Simplified way to create a vector layer displaying the countries of the world.
# Vector data is taken from folder data/ne_50m_admin_0_countries.geojson
# Input values to assign to countries come from a Pandas Dataframe df containing a column with iso2code and a column with values
# Returns a Map instance
###########################################################################################################################################################################
def countriesMap(df,                          # Pandas dataframe indexed on iso2_code and containing 'value' and 'label' columns
                 code_column=None,            # Name of the column containing the code of the country (None = the country code is the index of the dataframe)
                 value_column='value',        # Name of the column containing the value
                 codes_selected=[],           # codes of the countries selected
                 center=None,                 # [lat,lon] to center the map
                 zoom=None,                   # initial zoom level for the map
                 width ='99%',                # width of the map
                 height='400px',              # height of the map
                 min_width=None,              # min_width of the map
                 basemap=basemaps.OpenStreetMap.Mapnik, # Basemap to use
                 detailedcountries=False,     # If True loads the detailed country geojson
                 colorlist=['#0d0887', '#46039f', '#7201a8', '#9c179e', '#bd3786', '#d8576b', '#ed7953', '#fb9f3a', '#fdca26', '#f0f921'],   # default color scale
                 stdevnumber=2.0,             # Number of stddev to calculate (minvalue,maxvalue) range
                 stroke='#232323',            # stroke color for countries border
                 stroke_selected='#00ffff',   # stroke color for border of selected country
                 stroke_width=3.0,            # border width for countries polygons
                 decimals=2,                  # Number of decimals for the legend number display
                 minallowed_value=None,       # Minimum value allowed
                 maxallowed_value=None,       # Maximum value allowed
                 style      ={'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.6},    # Style to apply to the features
                 hover_style={'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.85}):  # Style to apply to the features when hover
    """
    Creation of an interactive map to display the countries of the world. An input Pandas DataFrame df is used to join a column of numeric values to the countries, using the iso2code (ISO 3166-2) as internal key attribute. Once the values are assigned to the countries, a graduated legend is calculated based on mean and standard deviation of the assigned values. A input list of colors is used to represent the countries given their assigned value.

    Parameters
    ----------
    df : Pandas DataFrame
        Pandas DataFrame to use for assigning values to the countries. It has to contain at least a column with numeric values.
    code_column : str, optional
        Name of the column of the Pandas DataFrame containing the unique code of the countries in the ISO-3166-2 standard. This column is used to perform the join with the internal attribute of the countries vector dataset that contains the country code. If the code_column is None, the code is taken from the index of the DataFrame, (default is None)
    value_column : str, optional
        Name of the column of the Pandas DataFrame containing the values to be assigned to the countries using the join on the ISO-3166-2 codes (default is 'value')
    codes_selected : list of strings, optional
        List of codes of countries to display as selected (default is [])
    center : tuple of (lat,lon), optional
        Geographical coordinates of the initial center of the interactive map visualization (default is None)
    zoom : int, optional
        Initial zoom level of the interactive map (default is None)
    width : str, optional
        Width of the map widget to create (default is '99%')
    height : str, optional
        Height of the map widget to create (default is '400px')
    min_width : str, optional
        Minimum width of the layout of the map widget (default is None)
    basemap : instance of basemaps type, optional
        Basemap to use as background map (default is basemaps.OpenStreetMap.Mapnik)
    detailedcountries : bool, optional
        If True loads the more detailed version of the countries dataset (default is False()
    colorlist : list of colors, optional
        List of colors to assign to the country polygons (default is the Plotly px.colors.sequential.Plasma, see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )
    stdevnumber : float, optional
        The correspondance between the values assigned to country polygons and the colors list is done by calculating a range of values [min,max] to linearly map the values to the colors. This range is defined by calculating the mean and standard deviation of the country values and applying this formula [mean - stdevnumber*stddev, mean + stdevnumber*stddev]. Default is 2.0
    stroke : str, optional
        Color to use for the border of countries (default is '#232323')
    stroke_selected : str, optional
        Color to use for the border of the selected countries (default is '#00ffff')
    stroke_width: float, optional
        Width of the border of the country polygons in pixels (default is 3.0)
    decimals : int, optional
        Number of decimals for the legend numbers display (default is 2)
    minallowed_value : float, optional
        Minimum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    maxallowed_value : float, optional
        Maximum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    style : dict, optional
        Style to apply to the features (default is {'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.6})
    hover_style : dict, optional
        Style to apply to the features when hover (default is {'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.85})
        
    Returns
    -------
        a ipyleaflet.Map instance

    Example
    -------
    Creation of a map displaying a random variable on 4 european countries. The numerical values assigned to each of the countries are randomly generated using numpy.random.uniform and saved into a dictionary having the country code as the key. This dict is transformed to a Pandas DataFrame with 4 rows and having 'iso2code' and 'value' as columns. The graduated legend is build using the 'inverted' Reds Plotly colorscale (low values are dark red, intermediate values are red, high values are white)::
        
        import numpy as np
        import pandas as pd
        import plotly.express as px
        from ipyleaflet import basemaps
        from vois import leafletMap

        countries = ['DE', 'ES', 'FR', 'IT']

        # Generate random values and create a dictionary: key=countrycode, value=random in [0.0,100.0]
        d = dict(zip(countries, list(np.random.uniform(size=len(countries),low=0.0,high=100.0))))

        # Create a pandas dataframe from the dictionary
        df = pd.DataFrame(d.items(), columns=['iso2code', 'value'])

        m = leafletMap.countriesMap(df,
                                    code_column='iso2code',
                                    height='400px',
                                    stroke_width=2.0,
                                    stroke_selected='yellow',
                                    basemap=basemaps.Stamen.Terrain,
                                    colorlist=px.colors.sequential.Reds[::-1],
                                    codes_selected=['IT'],
                                    center=[43,12], zoom=5)
        display(m)
        
    .. figure:: figures/leafletCountries.png
       :scale: 100 %
       :alt: leafletCountries example

       Example of an ipyleaflet Map displaying 4 european countries.
        
    """

    if df.shape[0] <= 0:
        minvalue = 1.0
        maxvalue = 2.0
    else:
        mean = df[value_column].mean()
        if df.shape[0] <= 1:
            minvalue = mean
            maxvalue = mean
        else:
            stddev = df[value_column].std()
            valuemin = df[value_column].min()
            valuemax = df[value_column].max()

            minvalue = mean - stdevnumber*stddev
            maxvalue = mean + stdevnumber*stddev

            if minvalue < valuemin: minvalue = valuemin
            if maxvalue > valuemax: maxvalue = valuemax

        if not minallowed_value is None:
            if minvalue < minallowed_value: minvalue = minallowed_value
        if not maxallowed_value is None:
            if maxvalue > maxallowed_value: maxvalue = maxallowed_value
        
    #print(mean,stddev, minvalue, maxvalue)

    if minvalue >= maxvalue: maxvalue = minvalue + 1
    ci = colors.colorInterpolator(colorlist,minvalue,maxvalue)

    
    # Assign colors for a feature
    def interpolate_color(feature):
        code  = feature['properties']['ISO_A2_EH']
        value = feature['properties']['value']
        if code in codes_selected:
            return { 'color': stroke_selected, 'weight': stroke_width+2, 'fillColor': ci.GetColor(value) }
        else:
            return { 'color': stroke,          'weight': stroke_width, 'fillColor': ci.GetColor(value) }

    
    # Creation of the Map
    m = Map(layout=Layout(width=width, height=height), scroll_wheel_zoom=True, basemap=basemap)
    if not min_width is None:
        m.layout.min_width = min_width
    if not center is None:
        m.center = center
    if not zoom is None:
        m.zoom = zoom

    # Add map controls
    m.add_control(FullScreenControl(position="topleft"))
    m.add_control(SearchControl(position="topleft",url='https://nominatim.openstreetmap.org/search?format=json&q={s}',zoom=12))
    m.add_control(ScaleControl(position='bottomright'))

    poslabel = widgets.Label(value='')
    widget_coordinate = WidgetControl(widget=poslabel, position='topright')
    m.add_control(widget_coordinate)
    
    lat = lon = 0
    def handle_interaction(**kwargs):
        nonlocal lat,lon
        if kwargs.get('type') == 'mousemove':
            lat = kwargs.get('coordinates')[0]
            lon = kwargs.get('coordinates')[1]
            poslabel.value = '{:.{prec}f}'.format(lat, prec=4) + ' - ' + '{:.{prec}f}'.format(lon, prec=4)

    m.on_interaction(handle_interaction)

   
    # Manage click on a feature
    def click_on_a_feature(*args, **kvargs):
        event   = kvargs['event']
        feature = kvargs['feature']

        code  = feature['properties']['ISO_A2_EH']
        name  = feature['properties']['NAME']
        value = feature['properties']['value']

        val = '{:.{prec}f}'.format(value, prec=decimals)
        s = name + ': ' + val
        
        pos = [lat,lon]
        message = widgets.HTML()
        message.value = "<style> p.small {line-height: 1.2; }</style><p class=\"small\">" + s + "</p>"
        popup = Popup(location=pos,child=message, close_button=True,auto_close=True,close_on_escape_key=True)
        m.add_layer(popup)
        
        
    # Add countries (taken from the setupfolder/data )
    path = Path(geojsonUtils.__file__)
    datafolder = str(path.parent.absolute()) + '/data'

    filepath = datafolder + '/ne_110m_admin_0_countries.geojson'
    if detailedcountries:
        filepath = datafolder + '/ne_50m_admin_0_countries.geojson'
    
    geojsonstr = geojsonUtils.geojsonLoadFile(filepath)

    countries = [str(x) for x in list(df[code_column])]
    values    = list(df[value_column])
    d = dict(zip(countries,values))
    geojsonnew = geojsonUtils.geojsonJoin(geojsonstr, 'ISO_A2_EH', 'value', d, innerMode=True)
    
    data = json.loads(geojsonnew)

    geo_json = GeoJSON(data=data, style=style, hover_style=hover_style, style_callback=interpolate_color)
    geo_json.on_click(click_on_a_feature)
    
    m.add_layer(geo_json)
    return m
    

    
###########################################################################################################################################################################
# Simplified way to create a vector layer displaying a custom geojson.
# Input values to assign to features come from a Pandas Dataframe df containing a column with unique code and a column with values
# Returns a Map instance
###########################################################################################################################################################################
def geojsonMap(df,                          # Pandas dataframe containing 'value' and 'label' columns
               geojson_path,                # Path of the geojson containing the geographic features
               geojson_attribute,           # Name of the attribute of the geojson containing the unique code
               code_column=None,            # Name of the column containing the code of the features (None = the feature code is the index of the dataframe)
               value_column='value',        # Name of the column containing the value
               codes_selected=[],           # codes of the features selected
               center=None,                 # [lat,lon] to center the map
               zoom=None,                   # initial zoom level for the map
               width ='99%',                # width of the map
               height='400px',              # height of the map
               min_width=None,              # min_width of the map
               basemap=basemaps.OpenStreetMap.Mapnik, # Basemap to use
               colorlist=['#0d0887', '#46039f', '#7201a8', '#9c179e', '#bd3786', '#d8576b', '#ed7953', '#fb9f3a', '#fdca26', '#f0f921'],   # default color scale
               stdevnumber=2.0,             # Number of stddev to calculate (minvalue,maxvalue) range
               stroke='#232323',            # stroke color for polygons border
               stroke_selected='#00ffff',   # stroke color for border of selected polygons
               stroke_width=3.0,            # border width for polygons
               decimals=2,                  # Number of decimals for the legend number display
               minallowed_value=None,       # Minimum value allowed
               maxallowed_value=None,       # Maximum value allowed
               style      ={'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.6},    # Style to apply to the features
               hover_style={'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.85}):  # Style to apply to the features when hover
    """
    Creation of an interactive map to display a custom geojson dataset. An input Pandas DataFrame df is used to join a column of numeric values to the geojson features, using the <geojson_attribute> as the internal key attribute. Once the values are assigned to the features, a graduated legend is calculated based on mean and standard deviation of the assigned values. A input list of colors is used to represent the featuress given their assigned value.

    Parameters
    ----------
    df : Pandas DataFrame
        Pandas DataFrame to use for assigning values to features. It has to contain at least a column with numeric values.
    geojson_path : str
        Path of the geojson file to load that contains the geographic features in geojson format
    geojson_attribute : str
        Name of the attribute of the geojson dataset that contains the unique codes of the features. This attribute will be use as internal key in the join operation with the df Pandas DataFrame
    code_column : str, optional
        Name of the column of the df Pandas DataFrame containing the unique code of the features. This column is used to perform the join with the internal attribute of the geojson vector dataset that contains the unique code. If the code_column is None, the code is taken from the index of the DataFrame, (default is None)
    value_column : str, optional
        Name of the column of the Pandas DataFrame containing the values to be assigned to the features using the join on geojson unique codes (default is 'value')
    codes_selected : list of strings, optional
        List of codes of features to display as selected (default is [])
    center : tuple of (lat,lon), optional
        Geographical coordinates of the initial center of the interactive map visualization (default is None)
    zoom : int, optional
        Initial zoom level of the interactive map (default is None)
    width : str, optional
        Width of the map widget to create (default is '99%')
    height : str, optional
        Height of the map widget to create (default is '400px')
    min_width : str, optional
        Minimum width of the layout of the map widget (default is None)
    basemap : basemap instance, optional
        Basemap to use as background in the map visualization (default is basemaps.OpenStreetMap.Mapnik). See `Documentation of ipyleaflet <https://ipyleaflet.readthedocs.io/en/latest/map_and_basemaps/basemaps.html>`_ for details
    colorlist : list of colors, optional
        List of colors to assign to the country polygons (default is the Plotly px.colors.sequential.Plasma, see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )
    stdevnumber : float, optional
        The correspondance between the values assigned to features and the colors list is done by calculating a range of values [min,max] to linearly map the values to the colors. This range is defined by calculating the mean and standard deviation of the country values and applying this formula [mean - stdevnumber*stddev, mean + stdevnumber*stddev]. Default is 2.0
    stroke : str, optional
        Color to use for the border of polygons (default is '#232323')
    stroke_selected : str, optional
        Color to use for the border of the selected polygons (default is '#00ffff')
    stroke_width: float, optional
        Width of the border of the polygons in pixels (default is 3.0)
    decimals : int, optional
        Number of decimals for the legend numbers display (default is 2)
    minallowed_value : float, optional
        Minimum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    maxallowed_value : float, optional
        Maximum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    style : dict, optional
        Style to apply to the features (default is {'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.6})
    hover_style : dict, optional
        Style to apply to the features when hover (default is {'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.85})
        
    Returns
    -------
        a ipyleaflet.Map instance

    Example
    -------
    Creation of a map displaying a custom geojson. The numerical values assigned to each of the countries are randomly generated using numpy.random.uniform and saved into a dictionary having the country code as the key. This dict is transformed to a Pandas DataFrame with 4 rows and having 'iso2code' and 'value' as columns. The graduated legend is build using the 'inverted' Viridis Plotly colorscale::
        
        import numpy as np
        import pandas as pd
        import plotly.express as px
        from vois import leafletMap

        countries = ['DE', 'ES', 'FR', 'IT']

        # Generate random values and create a dictionary: key=countrycode, value=random in [0.0,100.0]
        d = dict(zip(countries, list(np.random.uniform(size=len(countries),low=0.0,high=100.0))))

        # Create a pandas dataframe from the dictionary
        df = pd.DataFrame(d.items(), columns=['iso2code', 'value'])

        m = leafletMap.geojsonMap(df,
                                  './data/ne_50m_admin_0_countries.geojson',
                                  'ISO_A2_EH',   # Internal attribute used as key
                                  code_column='iso2code',
                                  height='400px',
                                  stroke_width=1.5,
                                  stroke_selected='yellow',
                                  colorlist=px.colors.sequential.Viridis[::-1],
                                  codes_selected=['IT'],
                                  center=[43,12], zoom=5)
        display(m)

    .. figure:: figures/leafletMap.png
       :scale: 100 %
       :alt: leafletMap example

       Example of an ipyleaflet Map displaying 4 european countries from a custom geojson file.
        
    """

    if df.shape[0] <= 0:
        minvalue = 1.0
        maxvalue = 2.0
    else:
        mean = df[value_column].mean()
        if df.shape[0] <= 1:
            minvalue = mean
            maxvalue = mean
        else:
            stddev = df[value_column].std()
            valuemin = df[value_column].min()
            valuemax = df[value_column].max()

            minvalue = mean - stdevnumber*stddev
            maxvalue = mean + stdevnumber*stddev

            if minvalue < valuemin: minvalue = valuemin
            if maxvalue > valuemax: maxvalue = valuemax

        if not minallowed_value is None:
            if minvalue < minallowed_value: minvalue = minallowed_value
        if not maxallowed_value is None:
            if maxvalue > maxallowed_value: maxvalue = maxallowed_value
        
    #print(mean,stddev, minvalue, maxvalue)

    if minvalue >= maxvalue: maxvalue = minvalue + 1
    ci = colors.colorInterpolator(colorlist,minvalue,maxvalue)

    
    # Assign colors for a feature
    def interpolate_color(feature):
        code  = feature['properties'][geojson_attribute]
        value = feature['properties']['value']
        if code in codes_selected:
            return { 'color': stroke_selected, 'weight': stroke_width+2, 'fillColor': ci.GetColor(value) }
        else:
            return { 'color': stroke,          'weight': stroke_width, 'fillColor': ci.GetColor(value) }

    
    # Creation of the Map
    m = Map(layout=Layout(width=width, height=height), scroll_wheel_zoom=True, basemap=basemap)
    if not min_width is None:
        m.layout.min_width = min_width
    if not center is None:
        m.center = center
    if not zoom is None:
        m.zoom = zoom

    # Add map controls
    m.add_control(FullScreenControl(position="topleft"))
    m.add_control(SearchControl(position="topleft",url='https://nominatim.openstreetmap.org/search?format=json&q={s}',zoom=12))
    m.add_control(ScaleControl(position='bottomright'))

    poslabel = widgets.HTML(value='')
    widget_coordinate = WidgetControl(widget=poslabel, position='topright')
    m.add_control(widget_coordinate)
    
    lat = lon = 0
    def handle_interaction(**kwargs):
        nonlocal lat,lon
        if kwargs.get('type') == 'mousemove':
            lat = kwargs.get('coordinates')[0]
            lon = kwargs.get('coordinates')[1]
            poslabel.value = '{:.{prec}f}'.format(lat, prec=4) + ' - ' + '{:.{prec}f}'.format(lon, prec=4)

    m.on_interaction(handle_interaction)

   
    # Manage click on a feature
    def click_on_a_feature(*args, **kvargs):
        event   = kvargs['event']
        feature = kvargs['feature']

        code  = feature['properties'][geojson_attribute]
        value = feature['properties']['value']

        val = '{:.{prec}f}'.format(value, prec=decimals)
        s = code + ': ' + val
        
        pos = [lat,lon]
        message = widgets.HTML()
        message.value = "<style> p.small {line-height: 1.2; }</style><p class=\"small\">" + s + "</p>"
        popup = Popup(location=pos,child=message, close_button=True,auto_close=True,close_on_escape_key=True)
        m.add_layer(popup)
        
        
    # Add layer
    geojsonstr = geojsonUtils.geojsonLoadFile(geojson_path)

    countries = [str(x) for x in list(df[code_column])]
    values    = list(df[value_column])
    d = dict(zip(countries,values))
    geojsonnew = geojsonUtils.geojsonJoin(geojsonstr, geojson_attribute, 'value', d, innerMode=True)
    
    data = json.loads(geojsonnew)

    geo_json = GeoJSON(data=data, style=style, hover_style=hover_style, style_callback=interpolate_color)
    geo_json.on_click(click_on_a_feature)
    
    m.add_layer(geo_json)
    return m


###########################################################################################################################################################################
# Simplified way to create a vector layer displaying a custom geojson with a categorical map on an attribute of the geojson
# Returns a Map instance
###########################################################################################################################################################################
def geojsonCategoricalMap(geojson_path,                # Path of the geojson containing the geographic features
                          geojson_attribute,           # Name of the attribute of the geojson containing the attribute to use for assigning the colors
                          center=None,                 # [lat,lon] to center the map
                          zoom=None,                   # initial zoom level for the map
                          width ='99%',                # width of the map
                          height='400px',              # height of the map
                          min_width=None,              # min_width of the map
                          basemap=basemaps.OpenStreetMap.Mapnik, # Basemap to use
                          colormap={},                 # dictionary containing geojson_attribute values as keys and colors as values
                          stroke='#232323',            # stroke color for polygons border
                          stroke_width=3.0,            # border width for polygons
                          fill='#aaaaaa',              # default fill color for polygons
                          style      ={'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.6},    # Style to apply to the features
                          hover_style={'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.85}):  # Style to apply to the features when hover
    """
    Creation of an interactive map to display a custom geojson dataset where colors are assigned to feature based on the values of an internal attribute of the input geojson file. The colormap parameter is a dictionary with keys corresponding to all the unique values of the internal attribute, which are mapped to the colors to use for representing each class.

    Parameters
    ----------
    geojson_path : str
        Path of the geojson file to load that contains the geographic features in geojson format
    geojson_attribute : str
        Name of the attribute of the geojson dataset that contains the thematisatio attribute of the features. This attribute will be used to retrieve the colors to assign to the features using the colormap parameter
    center : tuple of (lat,lon), optional
        Geographical coordinates of the initial center of the interactive map visualization (default is None)
    zoom : int, optional
        Initial zoom level of the interactive map (default is None)
    width : str, optional
        Width of the map widget to create (default is '99%')
    height : str, optional
        Height of the map widget to create (default is '400px')
    min_width : str, optional
        Minimum width of the layout of the map widget (default is None)
    basemap : basemap instance, optional
        Basemap to use as background in the map visualization (default is basemaps.OpenStreetMap.Mapnik). See `Documentation of ipyleaflet <https://ipyleaflet.readthedocs.io/en/latest/map_and_basemaps/basemaps.html>`_ for details
    colormap : dictionary containing geojson_attribute values as keys and colors as values
        Colors to assign to each distinct value of the geojson_attribute
    stroke : str, optional
        Color to use for the border of polygons (default is '#232323')
    stroke_width: float, optional
        Width of the border of the polygons in pixels (default is 3.0)
    fill : str, optional
        Default fill color to use for the polygons (default is '#aaaaaa')
    style : dict, optional
        Style to apply to the features (default is {'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.6})
    hover_style : dict, optional
        Style to apply to the features when hover (default is {'opacity': 1, 'dashArray': '0', 'fillOpacity': 0.85})
        
    Returns
    -------
        a ipyleaflet.Map instance

    Example
    -------
    Creation of a map displaying a custom geojson with data on landuse. The colors are assigned to the polygons based on the values of the 'fclass' attribute of the input geojson file. The colormap parameter is a dictionary with keys corresponding to all the unique landuse classes, which are mapped to the colors of a Plotly discrete color scale (see `Color Sequences in Plotly Express <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_)::
        
        import plotly.express as px
        from IPython.display import display
        from ipywidgets import widgets, Layout
        from vois import leafletMap, svgUtils, geojsonUtils

        # Load landuse example and get unique landuse classes
        filepath = './data/landuse.geojson'
        geojson = geojsonUtils.geojsonLoadFile(filepath)
        landuses = sorted(list(set(geojsonUtils.geojsonAll(geojson,'fclass'))))

        # Create a colormap (dictionary that maps landuses to colors)
        colors   = px.colors.qualitative.Dark24
        colormap = dict(zip(landuses, colors))

        m = leafletMap.geojsonCategoricalMap(filepath,
                                             'fclass',
                                             stroke_width=1.0,
                                             stroke='black', 
                                             colormap=colormap,
                                             width='79%',
                                             height='700px',
                                             center=[51.005,13.6],
                                             zoom=12,
                                             basemap=basemaps.CartoDB.Positron,
                                             style={'opacity': 1, 'dashArray': '0', 'fillOpacity': 1})
        
        outlegend = widgets.Output(layout=Layout(width='230px',height='680px'))
        with outlegend:
            display(HTML(svgUtils.categoriesLegend("Landuse legend",
                                                   landuses,
                                                   colorlist=colors[:len(landuses)])))
        widgets.HBox([m,outlegend])
        
    .. figure:: figures/leafletMapCat.png
       :scale: 100 %
       :alt: leafletMap example

       Example of an ipyleaflet Map displaying 4 european countries from a custom geojson file.
        
    """

    # Assign colors for a feature
    def get_color(feature):
        code  = feature['properties'][geojson_attribute]
        if code in colormap:
            return { 'color': stroke, 'weight': stroke_width, 'fillColor': colormap[code] }
        else:
            return { 'color': stroke, 'weight': stroke_width, 'fillColor': fill }
    
    # Creation of the Map
    m = Map(layout=Layout(width=width, height=height), scroll_wheel_zoom=True, basemap=basemap)
    if not min_width is None:
        m.layout.min_width = min_width
    if not center is None:
        m.center = center
    if not zoom is None:
        m.zoom = zoom

    # Add map controls
    m.add_control(FullScreenControl(position="topleft"))
    m.add_control(SearchControl(position="topleft",url='https://nominatim.openstreetmap.org/search?format=json&q={s}',zoom=12))
    m.add_control(ScaleControl(position='bottomright'))

    poslabel = widgets.HTML(value='')
    widget_coordinate = WidgetControl(widget=poslabel, position='topright')
    m.add_control(widget_coordinate)
    
    lat = lon = 0
    def handle_interaction(**kwargs):
        nonlocal lat,lon
        if kwargs.get('type') == 'mousemove':
            lat = kwargs.get('coordinates')[0]
            lon = kwargs.get('coordinates')[1]
            poslabel.value = '{:.{prec}f}'.format(lat, prec=4) + ' - ' + '{:.{prec}f}'.format(lon, prec=4)

    m.on_interaction(handle_interaction)

   
    # Manage click on a feature
    def click_on_a_feature(*args, **kvargs):
        event   = kvargs['event']
        feature = kvargs['feature']

        code  = feature['properties'][geojson_attribute]
        
        pos = [lat,lon]
        message = widgets.HTML()
        message.value = "<style> p.small {line-height: 1.2; }</style><p class=\"small\">" + code + "</p>"
        popup = Popup(location=pos,child=message, close_button=True,auto_close=True,close_on_escape_key=True)
        m.add_layer(popup)
        
        
    # Add layer
    geojsonstr = geojsonUtils.geojsonLoadFile(geojson_path)
    data = json.loads(geojsonstr)

    geo_json = GeoJSON(data=data, style=style, hover_style=hover_style, style_callback=get_color)
    geo_json.on_click(click_on_a_feature)
    
    m.add_layer(geo_json)
    return m



"""Extended selection of a palette of different families (sequential, divergent, etc.)"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.


# DOCUMENTATION ON MAPNIK:
# https://get-map.org/mapnik-lost-manual/book.pdf
# https://github.com/mapnik/mapnik-reference/blob/gh-pages/3.0.20/reference.json#L1517

from vois.vuetify import palettePicker, selectSingle, switch, sliderFloat

import ipyvuetify as v
from ipywidgets import widgets
from IPython.display import display


# Utility: convert three integers to '#RRGGBB'
def RGB(r,g,b):
    return '#{:02X}{:02X}{:02X}'.format(r, g, b)
    
custompalettes = [
    { "name": "Greyscale", "colors": ['#000000', '#FFFFFF']},
    
    { "name": "Simple", "colors": ['#000000', '#FF0000', '#00FF00', '#0000FF', '#FFFF00', '#FF00FF', '#00FFFF', '#FFFFFF']},
    
    { "name": "Dem",    "colors": [RGB(255,255,170), RGB( 39,168, 39), RGB( 11,128, 64), RGB(255,255,  0), RGB(255,186,  3),
                                   RGB(158, 30,  2), RGB(110, 40, 10), RGB(138, 94, 66), RGB(255,255,255)]},
    
    { "name": "NDVI",   "colors": [RGB(120,69,25), RGB(255,178,74), RGB(255,237,166), RGB(173,232,94),
                                   RGB(135,181,64), RGB(3,156,0), RGB(1,100,0), RGB(1,80,0)]}
]

families = ['carto', 'cmocean', 'cyclical', 'diverging', 'plotlyjs', 'qualitative', 'sequential', 'custom']



# Selection of a family and of a palette
class palettePickerEx():
    """
    Advanced selection of a palette of colors managing all the palette families and the interpolate flag
        
    Parameters
    ----------
    family : str, optional
        Family of the palette, one of these values: ['carto', 'cmocean', 'cyclical', 'diverging', 'plotlyjs', 'qualitative', 'sequential', 'custom'] (default is 'sequential')
    value : str, optional
        Name of the initially selected palette (default is 'Viridis')
    interpolate : bool, optional
        If True the colors are displayed as interpolated (default is True)
    show_interpolate_switch : bool, optional
        If True an interpolate switch is shown to enable or disable the interpolated display of the selected palette (default is True)
    width : int, optional
        Width of the widget in pixels (default is 400)
    clearable : bool, optional
        If True the dwopdown list of palettes will allow for no selection (default is True)
    onchange : function, optional
        Python function to call when the user selects one of the palettes. The function will pass as parameters the list of colors and the interpolate flag (default is None)
    show_opacity_slider : bool, optional
        If True an slider to select opacity is shown (default is False)
    onchangeOpacity : function, optional
        Python function to call when the user changes the opacity. The function will as parameter the selected opacity in [0.0,1.0] (default is None)

    Examples
    --------
    Creation of a selection widget for the palettes managing all the families::
        
        from vois.vuetify import palettePickerEx
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def onchange(colors, interpolate):
            with output:
                print(colors, interpolate)

        p = palettePickerEx.palettePickerEx(onchange=onchange)

        display(p.draw())
        display(output)

    .. figure:: figures/palettePickerEx.png
       :scale: 100 %
       :alt: palettePicker widget

       Example of an extended palette picker managing all the palette families and the interpolate flag
    """
    
    # Initialization
    def __init__(self,
                 family='sequential',
                 value='Viridis',
                 interpolate=True,
                 show_interpolate_switch=True,
                 onchange=None,
                 width=400,
                 clearable=True,
                 show_opacity_slider=True,
                 onchangeOpacity=None):
        
        self.family      = family
        self.interpolate = interpolate
        self.onchange    = onchange
        self.width       = width
        
        self.show_interpolate_switch = show_interpolate_switch
        
        self.show_opacity_slider = show_opacity_slider
        self.onchangeOpacity     = onchangeOpacity

        self.p = None
        self.sel = selectSingle.selectSingle('Family:', families, selection=family, width=150, onchange=self.onchangeFamily, marginy=1, clearable=False)
        self.sw  = switch.switch(self.interpolate, "Interpolate", onchange=self.onchangeInterpolate)
        
        if not self.show_interpolate_switch and self.show_opacity_slider:
            self.op = sliderFloat.sliderFloat(1.0, text='Opacity:', minvalue=0.0, maxvalue=1.0, sliderwidth=self.width-128, onchange=self.onchangeOpacity)
            
        self.p = palettePicker.palettePicker(family=self.family, custompalettes=custompalettes, label='Palette:', clearable=clearable, width=self.width, height=26, onchange=self.onchangePalette)
        self.p.value = value

        self.spacer = v.Html(tag='div',children=[' '], style_='width: 10px;')

    
    # Draw the widget
    def draw(self):
        if self.show_interpolate_switch:
            r = widgets.HBox([self.sel.draw(), self.spacer, self.sw.draw()])
        else:
            if self.show_opacity_slider:
                r = widgets.HBox([self.sel.draw(), self.spacer, self.op.draw()])
            else:
                r = self.sel.draw()
        return widgets.VBox([r, self.p.draw()])


    # Selection of a palette
    def onchangePalette(self):
        if not self.p is None:
            if not self.onchange is None:
                self.onchange(self.p.colors, self.interpolate)

    # Changed the family
    def onchangeFamily(self):
        self.family = self.sel.value
        if self.family == 'carto' or self.family == 'qualitative':
            self.interpolate = False
            self.sw.value = self.interpolate
        else:
            self.interpolate = True
            self.sw.value = self.interpolate
        self.p.updatePalettes(self.family,self.interpolate)
        self.p.value = self.p.images[0]['name']
        if not self.onchange is None:
            self.onchange(self.p.colors, self.interpolate)

    # Changed the interpolation flag
    def onchangeInterpolate(self, flag):
        self.interpolate = flag
        value = self.p.value
        oldonchange = self.onchange
        self.onchange = None
        self.p.updatePalettes(self.family,self.interpolate)
        self.onchange = oldonchange
        self.p.value = value
        if not self.onchange is None:
            self.onchange(self.p.colors, self.interpolate)

            
            
    # familyname property
    @property
    def familyname(self):
        """
        Get/Set name of the selected family.
        
        Returns
        --------
        name : str
            Name of the currently selected family

        Example
        -------
        Set and then get the current palette family::
            
            picker.familyname = 'qualitative'
            print(picker.familyname)
        
        """
        return self.sel.value
        
    
    # Select one of the families
    @familyname.setter
    def familyname(self, name):
        self.sel.value = name

        
    # value property
    @property
    def value(self):
        """
        Get/Set name of the selected palette.
        
        Returns
        --------
        name : str
            Name of the currently selected palette

        Example
        -------
        Set and then get the current palette name::
            
            picker.value = 'Viridis'
            print(picker.value)
        
        """
        return self.p.value
        
    
    # Select one of the palette given its name
    @value.setter
    def value(self, name):
        self.p.value = name
            
            
    # colors property
    @property
    def colors(self):
        """
        Get the colors of the selected palette.
        
        Returns
        --------
        colorlist : list of strings in '#RRGGBB' format
            List of colors of the selected palette

        Example
        -------
        Get the selected palette colors::
            
            print(picker.colors)
        
        """
        return self.p.colors
    
    
    # opacity property
    @property
    def opacity(self):
        """
        Get/Set the opacity value.
        
        Returns
        --------
        opacity : float
            Current value of opacity in thenrange [0.0,1.0]

        Example
        -------
        Set and then get the opacity::
            
            picker.opacity = 0.5
            print(picker.opacity)
        
        """
        if self.show_opacity_slider:
            return self.op.value
        return 1.0
        
    
    # Set the opacity
    @opacity.setter
    def opacity(self, value):
        self.op.value = value
    
            


"""Simple card with title, subtitle and image."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
import traitlets


class card(v.VuetifyTemplate):
    """
    Simple card displaying title, subtitle and an image.
        
    Parameters
    ----------
    width : str, optional
        Width of the card (default is '400px')
    color : str, optional
        Background color of the card (default is 'white')
    dark : bool, optional
        If True the title and subtitle texts are displayed in white color, if False they are displayed in black (default is False)
    ripple : bool, optional
        If True the click on the card is highlighted (default is False)
    elevation : int, optional
        Elevation of the card over the background of the page (default is 3)
    title : str, optional
        Title of the card (default is 'Title')
    subtitle : str, optional
        Subtitle of the card (default is 'Subtitle')
    icon : str, optional
        URL of the image to display as an icon before the card title (default is '')
    iconsize : str, optional
        Size of the area where the icon is displayed (default is '32px')
    image : str, optional
        URL of the image to display in the right side of the card (default is '')
    imagesize : str, optional
        Size of the area where the image is displayed (default is '190px')
    on_click : function, optional
        Python function to call when the user clicks on the card. The function will receive no parameters. (default is None)
    argument : any, optional
        Argument to pass to the on_click python function (default is None)
    responsive : bool, optional
        If True, the font size is automatically changed according to the page size (default is False)
    fontsizemultiplier : float, optional
        Multiply factor for changing the standard size of the font used for title and subtitle (default is 1.0)
    backgroundimageurl : str, optional
        URL of the optional image to display as background of the card (default is '')
    tooltip : str, optional
        Text to display as tooltip of the whole card (default is '')
    titletooltip : str, optional
        Text to display as tooltip of the card title (default is '')
    focusedopacity : float, optional
        Opacity of the card background when the card is clicked (has focus). Default is 0.1
    textcolor : str, optional
        Color of the text (default is 'black')
    titleweight : int, optional
        Font weight for the title (default is 700)
    subtitleweight : int, optional
        Font weight for the subtitle (default is 400)

    Example
    -------
    Creation of a card to display text and an image::
        
        from vois.vuetify import card
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def on_click():
            with output:
                print('clicked!')

        subtitle = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi...'
        width = 600

        c = card.card(elevation=5, width='600px', title='Sample dataset', subtitle=subtitle,
                      image='https://cdn.vuetifyjs.com/images/cards/halcyon.png', on_click=on_click)

        display(c)
        display(output)

    .. figure:: figures/card.png
       :scale: 100 %
       :alt: card widget

       Example of a card with text and an image
    """
    
    width      = traitlets.Unicode('400px').tag(sync=True)
    height     = traitlets.Unicode('').tag(sync=True)
    color      = traitlets.Unicode('white').tag(sync=True)
    dark       = traitlets.Bool(False).tag(sync=True)
    ripple     = traitlets.Bool(False).tag(sync=True)       # Ripple flag (if True the click on the card is highlighted)
    disabled   = traitlets.Bool(False).tag(sync=True)
    elevation  = traitlets.Int(5).tag(sync=True)
    title      = traitlets.Unicode('Title').tag(sync=True)
    subtitle   = traitlets.Unicode('Subtitle').tag(sync=True)
    icon       = traitlets.Unicode('').tag(sync=True)
    iconsize   = traitlets.Unicode('32px').tag(sync=True)
    image      = traitlets.Unicode('').tag(sync=True)
    imagesize  = traitlets.Unicode('190px').tag(sync=True)
    on_click   = traitlets.Any(None).tag(sync=False)
    argument   = traitlets.Any(None).tag(sync=False)
    responsive = traitlets.Bool(False).tag(sync=True)
    fontsizemultiplier = traitlets.Float(1.0).tag(sync=True)
    backgroundimageurl = traitlets.Unicode('').tag(sync=True)
    
    subtitlemargins = traitlets.Unicode('ma-0 ml-4 mb-4 mt-0 mr-4').tag(sync=True)
    focusedopacity  = traitlets.Float(0.1).tag(sync=True)
    textcolor       = traitlets.Unicode('black').tag(sync=True)
    titleweight     = traitlets.Int(700).tag(sync=True)
    subtitleweight  = traitlets.Int(400).tag(sync=True)

    
    @traitlets.default('template')
    def _template(self):
        
        pre  = ''
        ttip = ''
        post = ''
        
        if len(self.tooltip) > 0:
            pre  = '<v-hover v-slot="{ hover }">'
            ttip = '<v-tooltip attach allow-overflow absolute center bottom z-index=99999 v-model="hover"><span>%s</span></v-tooltip>' % self.tooltip
            post = '</v-hover>'

        title_pre  = ''
        title_att  = ''
        title_post = ''
        
        if len(self.titletooltip) > 0:
            title_pre  = '<v-tooltip bottom max-width="%s"><template v-slot:activator="{ on, attrs }">' % str(self.width)
            title_att  = 'v-bind="attrs" v-on="on"'
            title_post = '</template><span>%s</span></v-tooltip>' % self.titletooltip
                  

        return '''
%s

  <v-card
    class="pa-1 ma-1"
    :width="width"
    :height="height"
    :color="color"
    :dark="dark"
    raised
    hover
    :elevation="elevation"
    :ripple="ripple"
    :disabled="disabled"
    @click="clicked"
    style="overflow: hidden;"
    :img="backgroundimageurl"
  >
      <div class="d-flex flex-no-wrap justify-space-between">
        %s
        <div>
            <v-row dense>
                <v-avatar class="pa-0 ma-0 ml-3 mt-2 mr-n2" :size="iconsize" tile >
                    <v-img :src="icon"></v-img>
                </v-avatar>
                
                %s
                <v-card-title %s class="mt-n2 mb-1" :style="fontSizeTitle" v-text="title"></v-card-title>
                %s
            </v-row>
          <div :class="subtitlemargins" :style="fontSizeSubTitle" v-html="subtitle"/>
      </div>
      <v-avatar class="ma-n1" :size="imagesize" tile >
         <v-img :src="image" contain></v-img>
      </v-avatar>
    </div>
  </v-card>
%s

<script>
  export default {
    computed: {
      fontSizeTitle ()
        {
         if( this.responsive )
           {
            switch( this.$vuetify.breakpoint.name )
              {
               case 'xs': return 'color: %s; font-weight: %d; font-size: ' + (0.75*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'sm': return 'color: %s; font-weight: %d; font-size: ' + (1.00*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'md': return 'color: %s; font-weight: %d; font-size: ' + (1.25*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'lg': return 'color: %s; font-weight: %d; font-size: ' + (1.45*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'xl': return 'color: %s; font-weight: %d; font-size: ' + (1.60*this.fontsizemultiplier).toFixed(3) + 'em;'
              }
           }
         else
           {
            return 'colors: %s; font-weight: %d; font-size: ' + (1.25*this.fontsizemultiplier).toFixed(3) + 'em;'
           }
      },
        
      fontSizeSubTitle ()
        {
         if( this.responsive )
           {
            switch( this.$vuetify.breakpoint.name )
              {
               case 'xs': return 'color: %s; font-weight: %d; font-size: ' + (0.8*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'sm': return 'color: %s; font-weight: %d; font-size: ' + (0.9*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'md': return 'color: %s; font-weight: %d; font-size: ' + (1.0*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'lg': return 'color: %s; font-weight: %d; font-size: ' + (1.1*this.fontsizemultiplier).toFixed(3) + 'em;'
               case 'xl': return 'color: %s; font-weight: %d; font-size: ' + (1.2*this.fontsizemultiplier).toFixed(3) + 'em;'
              }
           }
         else
           {
            return 'color: %s; font-weight: %d; font-size: ' + (1.0*this.fontsizemultiplier).toFixed(3) + 'em;'
           }
        },
    },
  }
</script>

<style>
.vuetify-styles .v-card--link:focus::before {
  opacity: %f;
}
</style>


''' % (pre,ttip, title_pre,title_att,title_post, post,
       self.textcolor,self.titleweight,    self.textcolor,self.titleweight,    self.textcolor,self.titleweight,    self.textcolor,self.titleweight,    self.textcolor,self.titleweight,    self.textcolor,self.titleweight, 
       self.textcolor,self.subtitleweight, self.textcolor,self.subtitleweight, self.textcolor,self.subtitleweight, self.textcolor,self.subtitleweight, self.textcolor,self.subtitleweight, self.textcolor,self.subtitleweight, 
       self.focusedopacity)
    
    def __init__(self,
                 *args,
                 width='400px',
                 height='',
                 color='white',
                 dark=False,
                 ripple=False,
                 disabled=False,
                 elevation=3,
                 title='Title',
                 subtitle='Subtitle',
                 icon='',
                 iconsize='32px',
                 image='',
                 imagesize='190px',
                 on_click=None,
                 argument=None,
                 responsive=False,
                 fontsizemultiplier=1.0,
                 backgroundimageurl='',
                 tooltip='',
                 titletooltip='',
                 focusedopacity=0.1,     # Opacity for the background when the card is clicked (has focus)
                 textcolor='black',
                 titleweight=700,
                 subtitleweight=400,
                 **kwargs):

        self.tooltip        = tooltip
        self.titletooltip   = titletooltip
        self.focusedopacity = focusedopacity
        self.textcolor      = textcolor
        self.titleweight    = titleweight
        self.subtitleweight = subtitleweight
 
        self.width     = width
        self.height    = height
        self.color     = color
        self.dark      = dark
        self.ripple    = ripple
        self.disabled  = disabled
        self.elevation = elevation
        self.title     = title
        self.subtitle  = subtitle
        if len(self.subtitle) == 0:
            self.subtitlemargins = 'ma-0 ml-4 mb-n4 mt-0'

        self.icon = icon
        if len(icon) == 0: self.iconsize = '0px'
        else:              self.iconsize = iconsize
        
        self.image = image
        if len(image) == 0: self.imagesize = '0px'
        else:               self.imagesize = imagesize
        
        self.on_click = on_click
        self.argument = argument
        
        self.responsive         = responsive
        self.fontsizemultiplier = fontsizemultiplier
        self.backgroundimageurl = backgroundimageurl
        
        super().__init__(*args, **kwargs)
    
    # Manage event "click"
    def vue_clicked(self, data):
        if not self.on_click is None:
            if self.argument is None:
                self.on_click()
            else:
                self.on_click(self.argument)


"""Read parameters passed in the URL of the Voila dashboard"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

def readParameters():
    """
    Read parameters passed in the URL of the Voilà dashboard.
        
    Returns
    --------
    parameters : dictionary
        Dictionary containing key-values for all the URL passed to the Voilà page
        
    Example
    -------
    Read URL parameters and get value of one of the parameters::
    
        from vois.vuetify import queryStrings
        
        parameters = queryStrings.readParameters()
        activetab = parameters.get('activetab', ['chart'])[0]
        print(activetab)
        
    """
    import os
    import json
    run_in_voila = True
    try:
        from voila.utils import get_query_string
    except:
        run_in_voila = False

    from urllib.parse import parse_qs

    
    if run_in_voila:
        query_string = get_query_string()   # New in Voilà 0.3.0
    else:
        query_string = os.environ.get('QUERY_STRING', '')

    parameters = parse_qs(query_string)
    return parameters



import ipyvuetify as v
import traitlets
import io
import os
import IPython
import copy
import asyncio
import nest_asyncio
import sys

chunk_listener_id = 0

nest_asyncio.apply()


def load_template(filename):
    with open(os.path.join(os.path.dirname(__file__), filename)) as f:
        return f.read()


async def process_messages():
    ipython = IPython.get_ipython()
    original_parent_ident = ipython.kernel._parent_ident
    original_parent_header = ipython.kernel._parent_header
    original_set_parent = ipython.set_parent

    def set_parent_sink(*args):
        pass

    try:
        ipython.set_parent = set_parent_sink
        while not ipython.kernel.msg_queue.empty():
            await ipython.kernel.do_one_iteration()
    finally:
        # reset parent header to original execute request
        ipython.kernel.set_parent(original_parent_ident, original_parent_header)  # for execution count indicator
        sys.stdout.parent_header = original_parent_header  # for print statements
        sys.stderr.parent_header = original_parent_header  # for errors
        ipython.display_pub.parent_header = original_parent_header  # for display()
        ipython.set_parent = original_set_parent  # for displaying execution result


class ClientSideFile(io.RawIOBase):

    def __init__(self, widget, file_index, timeout=30):
        global chunk_listener_id
        self.id = chunk_listener_id
        self.widget = widget
        self.version = widget.version
        self.file_index = file_index
        self.timeout = timeout
        self.valid = True
        self.offset = 0
        self.size = widget.file_info[file_index]['size']

        self.chunk_queue = []

        widget.chunk_listeners[self.id] = self

        chunk_listener_id += 1
        self.waits = 0

    def handle_chunk(self, content, buffer):
        content['buffer'] = buffer
        self.chunk_queue.append(content)

    def readable(self):
        return True

    def seekable(self):
        return True

    def seek(self, offset, whence=io.SEEK_SET):
        if whence == io.SEEK_SET:
            self.offset = offset
        elif whence == io.SEEK_CUR:
            self.offset = self.offset + offset
        elif whence == io.SEEK_END:
            self.offset = self.size + offset
        else:
            raise ValueError(f'whence {whence} invalid')

    def tell(self):
        return self.offset

    def readinto(self, buffer):
        if not self.valid:
            raise Exception('Invalid file state')
        mem = memoryview(buffer)

        remaining = max(0, self.size - self.offset)
        size = min(len(buffer), remaining)

        self.widget.send({
            'method': 'read',
            'args': [{
                'file_index': self.file_index,
                'offset': self.offset,
                'length': size,
                'id': self.id
            }]
        })

        sleep_interval = 0.01
        max_iterations = self.timeout / sleep_interval

        async def read_all():
            bytes_read = 0
            while bytes_read < size:
                iterations = 0
                while not self.chunk_queue:
                    iterations += 1

                    if self.version != self.widget.version:
                        self.valid = False
                        raise Exception('File changed')
                    if iterations > max_iterations:
                        self.valid = False
                        raise Exception('Timeout')

                    await asyncio.sleep(sleep_interval)
                    await process_messages()

                self.waits += iterations

                chunk = self.chunk_queue[0]
                chunk_size = chunk['length']

                mem[bytes_read:bytes_read + chunk_size] = chunk['buffer']

                self.chunk_queue.pop(0)
                bytes_read += chunk_size
                self.offset += chunk_size

                self.widget.update_stats(self.file_index, chunk_size)
                await process_messages()

        asyncio.run(read_all())
        return size

    def readall(self):
        return self.read(self.size - self.offset)


class FileInput(v.VuetifyTemplate):
    template = traitlets.Unicode(load_template('file_input.vue')).tag(sync=True)
    data = traitlets.Unicode('{myfiles: undefined}').tag(sync=True)

    file_info = traitlets.List().tag(sync=True)
    version = traitlets.Int(0).tag(sync=True)
    multiple = traitlets.Bool(False).tag(sync=True)
    disabled = traitlets.Bool(False).tag(sync=True)
    directory = traitlets.Bool(False).tag(sync=True)
    accept = traitlets.Unicode().tag(sync=True)
    total_progress = traitlets.Int(0).tag(sync=True)
    show_progress = traitlets.Bool(True).tag(sync=True)
    progress_indeterminate = traitlets.Bool(False).tag(sync=True)
    color = traitlets.Unicode('red').tag(sync=True)
    placeholder = traitlets.Unicode().tag(sync=True)
    label = traitlets.Unicode().tag(sync=True)

    total_progress_inner = 0
    total_size_inner = 0

    def __init__(self, **kwargs):
        self.chunk_listeners = {}
        self.stats = []
        super().__init__(**kwargs)

    @traitlets.observe('file_info')
    def _file_info_changed(self, _):
        self.version += 1
        self.reset_stats()

    def update_stats(self, file_index, bytes_read):
        self.stats[file_index] += bytes_read
        tot = sum(self.stats)
        percent = round((tot / self.total_size_inner) * 100)
        if percent != self.total_progress_inner:
            self.total_progress_inner = percent
            self.total_progress = percent

    def get_files(self, timeout=30):
        files = []
        for index, file in enumerate(self.file_info):
            file = copy.deepcopy(self.file_info[index])
            file['file_obj'] = ClientSideFile(self, index, timeout=timeout)
            files.append(file)
        return files

    def clear(self):
        self.reset_stats()
        self.send({
            'method': 'clear',
            'args': []
        })

    def reset_stats(self):
        self.stats = [0 for _ in self.file_info]
        self.total_progress = 0
        self.total_progress_inner = 0
        self.total_size_inner = sum([f['size'] for f in self.file_info])

    def vue_upload(self, content, buffers):
        listener_id = content['id']
        listener = self.chunk_listeners.get(listener_id)
        if listener:
            if listener.version != self.version:
                del self.chunk_listeners[listener_id]
            else:
                listener.handle_chunk(content, buffers[0])


__all__ = ['FileInput']